const {
  BaseKonnector,
  log,
  errors,
  addData,
  updateOrCreate,
  hydrateAndFilter,
} = require('cozy-konnector-libs')

const axios = require('axios').default
const moment = require('moment')
require('moment-timezone')

moment.locale('fr') // set the language
moment.tz.setDefault('Europe/Paris') // set the timezone

const Sentry = require('@sentry/node')
// eslint-disable-next-line
const Tracing = require('@sentry/tracing') // Needed for tracking performance in Sentry
const { version } = require('../package.json')
const { isDev } = require('./helpers/env')

const manualExecution = process.env.COZY_JOB_MANUAL_EXECUTION === 'true'

const startDate = manualExecution
  ? moment()
      .subtract(1, 'year')
      .format('MM/DD/YYYY')
  : moment()
      .subtract(3, 'year')
      .format('MM/DD/YYYY')

const endDate = moment().format('MM/DD/YYYY')
const rangeDate = {
  day: {
    doctype: 'com.grandlyon.egl.day',
    keys: ['year', 'month', 'day'],
  },
  month: {
    doctype: 'com.grandlyon.egl.month',
    keys: ['year', 'month'],
  },
  year: {
    doctype: 'com.grandlyon.egl.year',
    keys: ['year'],
  },
}

module.exports = new BaseKonnector(start)
/**
 * Sentry
 */
Sentry.init({
  dsn:
    'https://3f97baf46c2b44c2bd9e0c371abe3e05@grandlyon.errors.cozycloud.cc/2',

  // Set tracesSampleRate to 1.0 to capture 100%
  // of transactions for performance monitoring.
  // We recommend adjusting this value in production
  tracesSampleRate: 1.0,
  release: version,
  environment: isDev() ? 'development' : 'production',
  debug: isDev(),
  integrations: [
    // enable HTTP calls tracing
    new Sentry.Integrations.Http({ tracing: true }),
  ],
})

// The start function is run by the BaseKonnector instance only when it got all the account
// information (fields). When you run this connector yourself in 'standalone' mode or 'dev' mode,
// the account information come from ./konnector-dev-config.json file
async function start(fields, cozyParameters) {
  const transaction = Sentry.startTransaction({
    op: 'konnector',
    name: 'EGL Konnector',
  })
  transaction.startChild({ op: 'Konnector starting' })

  try {
    // Local debug data
    // const baseUrl = fields.eglBaseURL
    // const apiAuthKey = fields.eglAPIAuthKey
    const baseUrl = cozyParameters.secret.eglBaseURL
    const apiAuthKey = cozyParameters.secret.eglAPIAuthKey
    log('info', 'Authenticating ...')
    const response = await authenticate(
      fields.login,
      fields.password,
      baseUrl,
      apiAuthKey
    )
    log('info', 'Successfully logged in')

    const eglData = await getData(response, baseUrl, apiAuthKey)
    if (eglData) {
      log('debug', 'Process egl daily data')
      const filteredDocuments = await hydrateAndFilter(
        eglData,
        rangeDate.day.doctype,
        {
          keys: rangeDate.day.keys,
        }
      )
      await addData(filteredDocuments, rangeDate.day.doctype)

      log('debug', 'Aggregate egl monthly load data')
      const monthlyLoads = aggregateMonthlyLoad(eglData)
      log('debug', 'Store aggregated egl monthly load data')
      await updateOrCreate(
        monthlyLoads,
        rangeDate.month.doctype,
        rangeDate.month.keys
      )

      log('debug', 'Aggregate egl yearly load data')
      const yearlyLoads = aggregateYearlyLoad(monthlyLoads)
      log('debug', 'Store aggregated egl yearly load data')
      await updateOrCreate(
        yearlyLoads,
        rangeDate.year.doctype,
        rangeDate.year.keys
      )
    } else {
      log('debug', 'No data found')
    }
    transaction.setStatus(Tracing.SpanStatus.Ok)
    transaction.finish()
  } catch (error) {
    const errorMessage = `EGL konnector encountered an error. Response data: ${JSON.stringify(
      error.message
    )}`
    Sentry.captureMessage(errorMessage, {
      tags: {
        section: 'start',
      },
    })
    transaction.setStatus(Tracing.SpanStatus.Aborted)
    transaction.finish()
    await Sentry.flush()
    throw error
  }
}

function aggregateMonthlyLoad(data) {
  const monthlyLoad = {}

  for (const entry of data) {
    const { year, month, load } = entry
    const monthKey = `${year}-${month}`

    if (!monthlyLoad[monthKey]) {
      monthlyLoad[monthKey] = {
        load: load,
        year: year,
        month: month,
        day: 0,
        hour: 0,
        minute: 0,
        type: 'R',
      }
    } else {
      monthlyLoad[monthKey].load += load
    }
  }

  return Object.values(monthlyLoad)
}

function aggregateYearlyLoad(data) {
  const yearlyLoad = {}

  for (const entry of data) {
    const { year, load } = entry
    if (!yearlyLoad[year]) {
      yearlyLoad[year] = {
        load: load,
        year: year,
        month: 0,
        day: 0,
        hour: 0,
        minute: 0,
        type: 'R',
      }
    } else {
      yearlyLoad[year].load += load
    }
  }

  return Object.values(yearlyLoad)
}

/**
 * @typedef {Object} AuthResponse
 * @property {number} codeRetour
 * @property {string} libelleRetour
 * @property {AuthResult} resultatRetour
 */

/**
 * @typedef {Object} AuthResult
 * @property {number} num_abt
 * @property {string} token
 */

/**
 * Authenticates a user with the provided credentials and returns an authentication response.
 *
 * @param {number} login - The user's login.
 * @param {string} password - The user's password.
 * @param {string} baseUrl - The base URL for the authentication request.
 * @param {string} apiAuthKey - The API authentication key.
 *
 * @throws {Error} - Throws a Cozy error (VENDOR_DOWN or LOGIN_FAILED) in case of authentication failure.
 * @returns {Promise<AuthResponse>} - The authentication response containing a token.
 */
async function authenticate(login, password, baseUrl, apiAuthKey) {
  log('info', 'Authenticating ...')
  const authRequest = {
    method: 'post',
    url: baseUrl + '/connect.aspx',
    headers: {
      AuthKey: apiAuthKey,
      'Content-Type': 'application/x-www-form-urlencoded',
    },
    data: {
      login: login,
      pass: password,
    },
  }

  try {
    /** @type {AuthResponse} */
    const respData = (await axios(authRequest)).data

    if (respData.codeRetour === 100) {
      return respData
    }
    const errorMessage = `Authentication failed. Response data: ${respData.libelleRetour}`
    log('error', errorMessage)
    throw new Error(errors.VENDOR_DOWN)
  } catch (error) {
    log('error', error.response?.data)
    Sentry.captureException(error, {
      tags: {
        section: 'authenticate',
      },
      extra: {
        compte: login,
      },
    })
    if (error.response?.data.codeRetour === -4) {
      throw new Error(errors.LOGIN_FAILED)
    }
    throw new Error(errors.VENDOR_DOWN)
  }
}

/**
 * @typedef {Object} GetDataResponse
 * @property {number} codeRetour
 * @property {string} libelleRetour
 * @property {Array<Releve>} resultatRetour
 */

/**
 * @typedef {Object} Releve
 * @property {string} DateReleve
 * @property {string} TypeAgregat
 * @property {number} ValeurIndex
 */

/**
 * Retrieves data from a specified API using the provided response data and API configuration.
 *
 * @param {ApiResponse} response - The authentication response containing a valid token.
 * @param {string} baseUrl - The base URL for the data request.
 * @param {string} apiAuthKey - The API authentication key.
 *
 * @throws {Error} - Throws an error with an error code in case of data retrieval failure.
 * @returns {Promise<FormattedData>} - A promise that resolves to the retrieved and formatted data.
 */
async function getData(response, baseUrl, apiAuthKey) {
  const dataRequest = {
    method: 'post',
    url: baseUrl + '/getAllAgregatsByAbonnement.aspx',
    headers: {
      AuthKey: apiAuthKey,
      'Content-Type': 'application/x-www-form-urlencoded',
    },
    data: {
      token: response.resultatRetour.token,
      num_abt: response.resultatRetour.num_abt,
      date_debut: startDate,
      date_fin: endDate,
    },
  }

  try {
    /** @type {GetDataResponse} */
    const respData = (await axios(dataRequest)).data

    // Sort data by date
    respData.resultatRetour.sort(function(a, b) {
      return new Date(a.DateReleve) - new Date(b.DateReleve)
    })
    switch (respData.codeRetour) {
      case 100:
        return format(respData)
      case -2:
        log(
          'error',
          `Get data failed. codeRetour -2. ${respData.libelleRetour}`
        )
        throw errors.LOGIN_FAILED
      case -1:
        log(
          'error',
          `Get data failed. codeRetour -1. ${respData.libelleRetour}`
        )
        throw errors.VENDOR_DOWN
      default:
        log(
          'error',
          `Get data failed. ${respData.codeRetour}. ${respData.libelleRetour}`
        )
        throw errors.UNKNOWN_ERROR
    }
  } catch (error) {
    log('debug', error.message)
    Sentry.captureException(error, {
      tags: {
        section: 'getData',
      },
      extra: {
        start: startDate,
        end: endDate,
      },
    })
    if (axios.isAxiosError(error)) {
      throw new Error(errors.VENDOR_DOWN)
    }
    throw error
  }
}

/**
 * Represents a formatted data object after processing the getData response.
 *
 * @typedef {Object} FormattedData
 * @property {number} load - The processed load value.
 * @property {number} year - The year of the data point.
 * @property {number} month - The month of the data point.
 * @property {number} day - The day of the data point.
 * @property {number} hour - The hour of the data point (in this case, always 0).
 * @property {number} minute - The minute of the data point (in this case, always 0).
 * @property {string} type - The type of the data point.
 */

/**
 * Formats and processes data retrieved from the getData response.
 *
 * @param {GetDataResponse} response - The data response to be formatted.
 *
 * @throws {Error} - Throws an error with an error code in case of data processing failure.
 * @returns {Array<FormattedData>} - An array of formatted data objects.
 */
function format(response) {
  log('info', 'origin response size is: ' + response.resultatRetour.length)
  // Store first value as reference for index processing
  let refValue = response.resultatRetour[0]

  // Create copy of data without first value
  const data = response.resultatRetour
    .slice(1)
    .filter(value => value.ValeurIndex)

  log('info', 'filtered size is: ' + data.length)

  try {
    return data.map(value => {
      const time = moment(value.DateReleve, moment.ISO_8601)
      const processedLoad = value.ValeurIndex - refValue.ValeurIndex

      if (processedLoad < 0) {
        const errorMessage = `Processing load error for day ${parseInt(
          time.format('D')
        )}/${parseInt(time.format('M'))}/${parseInt(
          time.format('YYYY')
        )}, value is: ${processedLoad}`
        log('debug', errorMessage)
        throw errors.VENDOR_DOWN
      }

      // Change index ref value
      refValue = value

      return {
        load: processedLoad,
        year: parseInt(time.format('YYYY')),
        month: parseInt(time.format('M')),
        day: parseInt(time.format('D')),
        hour: 0,
        minute: 0,
        type: value.TypeAgregat,
      }
    })
  } catch (error) {
    log('debug', error.message)
    Sentry.captureException(error, {
      tags: {
        section: 'format',
      },
    })
    throw error
  }
}
